/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 * 
 * Copyright (C) 2004 - 2012 CASSIDIAN
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */

package org.ow2.weblab.webservices.listener;

import java.io.File;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.ow2.weblab.content.api.ContentManager;
import org.ow2.weblab.core.extended.exception.WebLabCheckedException;
import org.ow2.weblab.core.extended.util.ResourceUtil;
import org.ow2.weblab.core.model.Resource;
import org.ow2.weblab.core.services.AccessDeniedException;
import org.ow2.weblab.core.services.ContentNotAvailableException;
import org.ow2.weblab.core.services.EmptyQueueException;
import org.ow2.weblab.core.services.InsufficientResourcesException;
import org.ow2.weblab.core.services.InvalidParameterException;
import org.ow2.weblab.core.services.QueueManager;
import org.ow2.weblab.core.services.ServiceNotConfiguredException;
import org.ow2.weblab.core.services.UnexpectedException;
import org.ow2.weblab.core.services.UnsupportedRequestException;
import org.ow2.weblab.core.services.queuemanager.NextResourceArgs;
import org.ow2.weblab.core.services.queuemanager.NextResourceReturn;
import org.ow2.weblab.webservices.listener.bean.FolderListenerConfigBean;
import org.ow2.weblab.webservices.listener.tool.FolderListenerManager;

/**
 * This class is an abstraction of a component that listens to folders in a list of directory in order to process them and, at the end provide expose them as a
 * QueueManager WebLab interface.
 * 
 * This class is abstract. Various implementations are provided for any files, WebLab resources, WARC files...
 * 
 * @author Jérémie Doucy, Arnaud Saval, Yann Mombrun
 */
public abstract class AbstractFolderListener implements QueueManager {


	protected final ContentManager contentManager;


	/**
	 * The folderListener for this instance, per usageContext
	 */
	protected final Map<String, FolderListenerManager> flms;


	/**
	 * The configuration bean.
	 */
	protected FolderListenerConfigBean conf;


	protected final Log log;


	/**
	 * Construct the service using the configuration bean
	 */
	public AbstractFolderListener(final FolderListenerConfigBean configBean) {
		this.flms = new HashMap<String, FolderListenerManager>();
		this.log = LogFactory.getLog(this.getClass());
		this.contentManager = ContentManager.getInstance();

		if (!this.validateBean(configBean)) {
			this.log.error("Configuration is not valid.");
			throw new InstantiationError("Configuration is not valid.");
		}
		this.conf = configBean;


		final Set<String> toStart = new HashSet<String>();
		if ((configBean.getPreInitialisedUsageContext() != null) && !configBean.getPreInitialisedUsageContext().isEmpty()) {
			toStart.addAll(configBean.getPreInitialisedUsageContext());
		}
		if (configBean.getDefaultUsageContext() != null) {
			toStart.add(configBean.getDefaultUsageContext());
		}
		for (final String uc : toStart) {
			this.flms.put(uc, new FolderListenerManager(configBean));
		}

	}


	@Deprecated
	protected Resource nextResource() throws EmptyQueueException, UnexpectedException {
		this.log.warn("This method is no more supported!");
		final NextResourceArgs nra = new NextResourceArgs();
		nra.setUsageContext(this.conf.getDefaultUsageContext());
		final NextResourceReturn nrr;
		try {
			nrr = this.nextResource(nra);
		} catch (final EmptyQueueException eqe) {
			throw eqe;
		} catch (final Exception e) {
			throw this.createUnexpectedException("An error occured in a method which is no more supported...", e);
		}
		return nrr.getResource();
	}


	/**
	 * @param usageContext
	 *            The <code>usageContext</code> used. Passed for debugging purposes.
	 * @return The created instance of Resource.
	 * @throws UnexpectedException
	 *             If an error occurs creating the resource.
	 * @throws EmptyQueueException
	 *             If the <code>Queue</code> is empty for <code>usageContext</code>.
	 * @throws ServiceNotConfiguredException
	 *             If no <code>Queue</code> is mapped to <code>usageContext</code> and if the <code>failForNonInitialisedUsageContext</code> is
	 *             <code>true</code>.
	 */
	protected abstract Resource nextResource(final String usageContext) throws UnexpectedException, EmptyQueueException, ServiceNotConfiguredException;


	@Override
	public NextResourceReturn nextResource(final NextResourceArgs args) throws ContentNotAvailableException, EmptyQueueException,
			InsufficientResourcesException, UnexpectedException, InvalidParameterException, ServiceNotConfiguredException, UnsupportedRequestException,
			AccessDeniedException {

		final String usageContext = this.checkArgs(args);
		this.log.debug("NextResource method called with \"" + usageContext + "\" as usageContext.");

		final Resource resource = this.nextResource(usageContext);
		this.log.debug("Resource \"" + resource.getUri() + "\" created for usageContext \"" + usageContext + "\".");

		if (this.log.isTraceEnabled()) {
			try {
				this.log.trace(ResourceUtil.saveToXMLString(resource));
			} catch (final WebLabCheckedException wlce) {
				this.log.warn("Unable to serialise resource " + resource.getUri() + ".", wlce);
			}
		}

		final NextResourceReturn nextResourceReturn = new NextResourceReturn();
		nextResourceReturn.setResource(resource);
		return nextResourceReturn;
	}


	/**
	 * @param usageContext
	 *            The <code>UsageContext</code> used to get the <code>Queue</code> to poll from
	 * @return The <code>File</code> that was on the head of the <code>Queue</code> mapped for <code>usageContext</code>
	 * @throws ServiceNotConfiguredException
	 *             If no <code>Queue</code> is mapped to <code>usageContext</code> and if the <code>failForNonInitialisedUsageContext</code> is
	 *             <code>true</code>.
	 * @throws UnexpectedException
	 *             If an error occurs while trying to initialise a <code>Queue</code> is for <code>usageContext</code>.
	 * @throws EmptyQueueException
	 *             If the <code>Queue</code> is empty for <code>usageContext</code>.
	 */
	protected synchronized File nextFile(final String usageContext) throws ServiceNotConfiguredException, UnexpectedException, EmptyQueueException {
		final FolderListenerManager flm;
		if (!this.flms.containsKey(usageContext)) {
			if (this.conf.isFailForNonInitialisedUsageContext()) {
				final String message = "UsageContext \"" + usageContext + "\" not configured and lazy start not set.";
				this.log.error(message);
				throw new ServiceNotConfiguredException(message, message);
			}
			flm = new FolderListenerManager(this.conf);
			if (!flm.start()) {
				throw this.createUnexpectedException("UsageContext \"" + usageContext + "\" not configured. "
						+ "An error happend when initialising a new listener.", null);
			}
			this.flms.put(usageContext, flm);
		} else {
			flm = this.flms.get(usageContext);
		}
		if (flm.getFifo().isEmpty()) {
			throw this.noMoreResources(usageContext);
		}
		// Might return null but should not!
		final File file = flm.getFifo().poll();
		if (file == null) {
			throw this.noMoreResources(usageContext);
		}
		return file;
	}


	/**
	 * @param args
	 *            The <code>NextResourceArgs</code> received.
	 * @return The <code>usageContext</code> String in <code>args</code> or <code>null</code> if no <code>usageContext</code> is found.
	 * @throws InvalidParameterException
	 *             If <code>args</code> is <code>null</code>.
	 */
	protected String checkArgs(final NextResourceArgs args) throws InvalidParameterException {
		if (args == null) {
			final String message = "NextResourceArgs was null.";
			this.log.error(message);
			throw new InvalidParameterException(message, message);
		}
		String uc = args.getUsageContext();
		if ((uc == null) && (this.conf.getDefaultUsageContext() != null)) {
			this.log.debug("No usageContext found. Using default one.");
			uc = this.conf.getDefaultUsageContext();
		} else if (uc == null) {
			final String message = "Null UsageContext is not supported.";
			this.log.error(message);
			throw new InvalidParameterException(message, message);
		}
		return uc;
	}


	@PreDestroy
	public void destroy() {
		this.log.debug("Stopping each listeners.");
		for (final FolderListenerManager listeners : this.flms.values()) {
			listeners.stop();
		}
	}


	@PostConstruct
	public boolean launch() {
		boolean worked = true;
		for (final FolderListenerManager listeners : this.flms.values()) {
			worked &= listeners.start();
		}
		return worked;
	}


	/**
	 * @param usageContext
	 *            The <code>usageContext</code> which is empty
	 * @return EmptyQueueException The created <code>Exception</code> to be thrown
	 */
	protected EmptyQueueException noMoreResources(final String usageContext) {
		final String message = "No more resources for UsageContext \"" + usageContext + "\".";
		this.log.error(message);
		return new EmptyQueueException(message, message);
	}


	/**
	 * @param message
	 *            The message to be logged as error and used in the exception
	 * @param cause
	 *            The cause of the error. Might be null
	 * @return EmptyQueueException The created <code>Exception</code> to be thrown
	 */
	protected UnexpectedException createUnexpectedException(final String message, final Throwable cause) {
		if (cause == null) {
			this.log.error(message);
			return new UnexpectedException(message, message);
		}
		this.log.error(message, cause);
		return new UnexpectedException(message, message, cause);
	}


	protected boolean validateBean(final FolderListenerConfigBean bean) {
		if (bean == null) {
			this.log.error("Configuration bean was null.");
			return false;
		}

		if (bean.getFileUrlPattern() == null) {
			this.log.error("FileUrlPattern in Configuration bean was null.");
			return false;
		}

		try {
			Pattern.compile(bean.getFileUrlPattern());
		} catch (final PatternSyntaxException pse) {
			this.log.error("FileUrlPattern [" + bean.getFileUrlPattern() + "] in Configuration bean was not valid.");
			return false;
		}

		if (bean.getUpdateInterval() <= 0) {
			this.log.error("UpdateInterval [" + bean.getUpdateInterval() + "] in Configuration bean was null.");
			return false;
		}

		if (bean.getFolders() == null) {
			this.log.error("Folders in Configuration bean was null.");
			return false;
		}

		if (bean.getFolders().isEmpty()) {
			this.log.error("No Folders in Configuration bean was null.");
			return false;
		}
		for (final File folder : bean.getFolders()) {
			if (!folder.isAbsolute()) {
				this.log.warn("Folder " + folder.getPath() + " is not absolute. It will be dependent of the way the server is started.");
			}
			if (!folder.isDirectory()) {
				this.log.warn("Folder " + folder.getPath() + " is not a directory. An error may occur later...");
			}
		}

		final boolean preInit;
		if ((bean.getPreInitialisedUsageContext() == null) || bean.getPreInitialisedUsageContext().isEmpty()) {
			this.log.warn("No PreInitialisedUsageContext provided in the Configuration bean.");
			preInit = false;
		} else {
			preInit = true;
		}

		if ((bean.getDefaultUsageContext() == null) && bean.isFailForNonInitialisedUsageContext() && !preInit) {
			this.log.error("No PreInitialisedUsageContext, no DefaultUsageContext and FailForNonInitialisedUsageContext. This is not possible!");
			return false;
		}

		return true;
	}

}
