/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 * 
 * Copyright (C) 2004 - 2012 CASSIDIAN
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */

package org.ow2.weblab.webservices.listener;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Date;

import javax.jws.WebService;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.ow2.weblab.core.extended.exception.WebLabCheckedException;
import org.ow2.weblab.core.extended.factory.ResourceFactory;
import org.ow2.weblab.core.model.Document;
import org.ow2.weblab.core.model.Resource;
import org.ow2.weblab.core.model.processing.WProcessingAnnotator;
import org.ow2.weblab.core.services.EmptyQueueException;
import org.ow2.weblab.core.services.ServiceNotConfiguredException;
import org.ow2.weblab.core.services.UnexpectedException;
import org.ow2.weblab.webservices.listener.bean.FolderListenerConfigBean;
import org.purl.dc.elements.DublinCoreAnnotator;
import org.purl.dc.terms.DCTermsAnnotator;

/**
 * 
 * Default implementation of a folder listener, simply create a valid WL resource from a local URL.
 * 
 * @author Jérémie Doucy, Arnaud Saval, Yann Mombrun, CASSIDIAN, an EADS company
 */
@WebService(endpointInterface = "org.ow2.weblab.core.services.QueueManager")
public class DefaultFolderListener extends AbstractFolderListener {


	protected static final String CRAWLER_ID = "folder-listener";


	/**
	 * @see AbstractFolderListener.AbstractFolderListener(FolderListenerConfigBean)
	 */
	public DefaultFolderListener(final FolderListenerConfigBean listenerConfigBean) {
		super(listenerConfigBean);
	}


	/**
	 * @param file
	 *            The <code>File</code> that was at the head of the <code>Queue</code> for <code>usageContext</code>
	 * @param usageContext
	 *            The <code>usageContext</code> used. Passed for debugging purposes.
	 * @return The created instance of Resource.
	 * @throws UnexpectedException
	 *             If an error occurs creating the resource.
	 */
	protected Resource createResource(final File file, final String usageContext) throws UnexpectedException {
		final Document document = ResourceFactory.createResource(DefaultFolderListener.CRAWLER_ID, "file" + System.nanoTime(), Document.class);

		this.log.debug("Storing file: " + file.getAbsolutePath());

		final FileInputStream fis;
		try {
			fis = FileUtils.openInputStream(file);
		} catch (final IOException ioe) {
			throw this.createUnexpectedException("An error occurs accessing file [" + file.getAbsolutePath() + "].", ioe);
		}

		try {
			this.contentManager.writeNativeContent(fis, document);
		} catch (final WebLabCheckedException wlce) {
			throw this.createUnexpectedException("An error occurs saving file [" + file.getAbsolutePath() + "].", wlce);
		}

		String path;
		try {
			path = file.getCanonicalPath();
		} catch (final IOException ioe) {
			this.log.warn("Unable to get canonical path of file: " + file.getAbsolutePath() + "; absolute path will be used instead.");
			path = file.getAbsolutePath();
		}
		path = FilenameUtils.normalize(path);

		final String filename = FilenameUtils.getName(file.getName());
		final long size = file.length();

		// Add WebLab Processing Annotations
		final WProcessingAnnotator wpa = new WProcessingAnnotator(document);
		wpa.writeGatheringDate(new Date());
		wpa.writeOriginalFileName(filename);
		wpa.writeOriginalFileSize(Long.valueOf(size));

		// Add Dublin Core Annotation
		final DublinCoreAnnotator dca = new DublinCoreAnnotator(document);
		dca.writeSource(path);

		// Add Dublin Core Terms Annotations
		final DCTermsAnnotator dcta = new DCTermsAnnotator(document);
		dcta.writeExtent(FileUtils.byteCountToDisplaySize(size));
		dcta.writeModified(new Date(file.lastModified()));

		return document;
	}


	@Override
	protected Resource nextResource(final String usageContext) throws UnexpectedException, EmptyQueueException, ServiceNotConfiguredException {
		final File file = this.nextFile(usageContext);
		this.log.debug("Next file for UsageContext \"" + usageContext + "\" is " + file.getAbsolutePath() + ".");

		final Resource resource = this.createResource(file, usageContext);

		if (this.conf.isDelete()) {
			this.log.trace("Remove file " + file.getAbsolutePath() + ".");
			if (!FileUtils.deleteQuietly(file)) {
				this.log.warn("An error occurs deleting file " + file.getAbsolutePath() + ".");
			}
		}

		return resource;
	}

}
