/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 * 
 * Copyright (C) 2004 - 2012 Cassidian, an EADS company
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */

package org.ow2.weblab.webservices.listener;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import javax.jws.WebService;

import org.apache.commons.io.FileUtils;
import org.archive.io.ArchiveRecord;
import org.archive.io.warc.WARCReader;
import org.archive.io.warc.WARCReaderFactory;
import org.ow2.weblab.core.extended.exception.WebLabCheckedException;
import org.ow2.weblab.core.extended.factory.ResourceFactory;
import org.ow2.weblab.core.model.Document;
import org.ow2.weblab.core.model.Resource;
import org.ow2.weblab.core.model.processing.WProcessingAnnotator;
import org.ow2.weblab.core.services.EmptyQueueException;
import org.ow2.weblab.core.services.ServiceNotConfiguredException;
import org.ow2.weblab.core.services.UnexpectedException;
import org.ow2.weblab.webservices.listener.bean.FolderListenerConfigBean;
import org.ow2.weblab.webservices.listener.bean.WarcConfigBean;
import org.purl.dc.elements.DublinCoreAnnotator;

/**
 * An implementation of filder listener that process Warc archives.
 * Warcs are processed as files containers.
 * Once all files in a Warcs have been returned, we process the next warc.
 * When all warcs are processed, we throw an exception to tell that there is no more resources.
 * 
 * @author asaval, ymombrun, jdoucy
 */
@WebService(endpointInterface = "org.ow2.weblab.core.services.QueueManager")
public class WarcListener extends AbstractFolderListener {


	public static final String CONTENT_TYPE = ".*Content-Type: (\\S*).*";


	public static final String DEFAULT_ACCEPTED_MESSAGE = "(.*response.*)";


	public static final String DEFAULT_ACCEPTED_HEADERS = ".*HTTP\\S*\\s*2\\d{2}.*";


	public static final String DEFAULT_ACCEPTED_CONTENT_TYPE = "(.*html.*)|(.*postscript.*)|(.*octet-stream.*)|(.*excel.*)"
			+ "|(.*powerpoint.*)|(.*word.*)|(.*pdf.*)|(.*opendocument.*)|(.*star.*)";


	protected final Map<String, Iterator<ArchiveRecord>> archiveRecordIterator;


	protected final Map<String, WARCReader> warcReader;


	protected final Map<String, File> currentWarcFile;


	protected String exposedAsPrefix;


	/*
	 * group1 = content type
	 */
	protected final Pattern httpMimeTypeHeaderPattern;


	protected final Pattern acceptedMessagePattern;


	protected final Pattern acceptedHeaders;


	private final SimpleDateFormat simpleDateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");


	protected final Pattern acceptedContentTypePattern;


	public WarcListener(final WarcConfigBean warcConfigBean) {
		super(warcConfigBean);
		this.acceptedMessagePattern = Pattern.compile(warcConfigBean.getAcceptedMessagePattern());
		this.acceptedContentTypePattern = Pattern.compile(warcConfigBean.getAcceptedContentTypePattern());
		this.acceptedHeaders = Pattern.compile(warcConfigBean.getAcceptedHeaders(), Pattern.DOTALL);
		this.httpMimeTypeHeaderPattern = Pattern.compile(WarcListener.CONTENT_TYPE, Pattern.DOTALL);
		this.exposedAsPrefix = warcConfigBean.getExposedAsPrefix();
		this.archiveRecordIterator = new HashMap<String, Iterator<ArchiveRecord>>();
		this.warcReader = new HashMap<String, WARCReader>();
		this.currentWarcFile = new HashMap<String, File>();
	}



	@Override
	protected synchronized Resource nextResource(final String usageContext) throws UnexpectedException, EmptyQueueException, ServiceNotConfiguredException {
		Resource resource = null;
		while (resource == null) {
			try {
				resource = this.nextWLDocument(usageContext);
			} catch (final EmptyQueueException eqe) {
				// current warc is empty. We don't care about the exception since it is only the loadNextWarc() that launch the final exception
				this.loadNextWarc(usageContext);
				resource = this.nextResource(usageContext);
			}
		}
		return resource;
	}



	/**
	 * Try to load next warc among the files
	 * 
	 * @param usageContext
	 *            The usageContext
	 * @throws EmptyQueueException
	 * @throws UnexpectedException
	 * @throws ServiceNotConfiguredException
	 */
	protected void loadNextWarc(final String usageContext) throws EmptyQueueException, ServiceNotConfiguredException, UnexpectedException {
		// check if we must delete previous warc file. We delete only warc file, if there is other files in the folder, we do not delete them
		final File oldWarc = this.currentWarcFile.get(usageContext);
		if (this.conf.isDelete() && (oldWarc != null)) {
			this.log.info("Deleting " + oldWarc.getAbsolutePath());
			FileUtils.deleteQuietly(oldWarc);
		}

		boolean loaded = false;
		do {
			final File currentWarc = this.nextFile(usageContext);
			this.currentWarcFile.put(usageContext, currentWarc);
			if (currentWarc == null) {
				throw this.noMoreResources(usageContext);
			}
			try {
				final WARCReader currentReader = WARCReaderFactory.get(currentWarc.getAbsolutePath());
				this.warcReader.put(usageContext, currentReader);
				this.archiveRecordIterator.put(usageContext, currentReader.iterator());
				loaded = true;
			} catch (final MalformedURLException murle) {
				this.log.error(murle, murle);
			} catch (final IOException ioe) {
				this.log.error(ioe, ioe);
			}
		} while (!loaded);
	}



	public Document nextWLDocument(final String usageContext) throws EmptyQueueException {
		final Iterator<ArchiveRecord> currentIterator = this.archiveRecordIterator.get(usageContext);
		if (currentIterator == null) {
			this.log.debug("UsageContext \"" + usageContext + "\" not initialised.");
			// TODO Initialise the UC if enabled in the configuration.
			throw this.noMoreResources(usageContext);
		}

		ArchiveRecord archiveRecord = null;
		String contentType = null;
		Document document = null;
		while (currentIterator.hasNext() && (document == null)) {
			contentType = null;
			archiveRecord = currentIterator.next();
			final String mimetype = archiveRecord.getHeader().getMimetype();
			if (this.acceptedMessagePattern.matcher(mimetype).matches()) {
				String header = "";
				try {
					header = this.skipHttpHeader(archiveRecord);
				} catch (final IOException e1) {
					this.log.error(e1, e1);
				}
				if (this.acceptedHeaders.matcher(header).matches()) {
					final Matcher contentTypeMatcher = this.httpMimeTypeHeaderPattern.matcher(header);
					if (contentTypeMatcher.matches()) {
						contentType = contentTypeMatcher.group(1).trim();
						if (this.acceptedContentTypePattern.matcher(contentType).matches()) {
							document = this.generateDocumentWithNative(archiveRecord);
						}
					}
				}
			}
		}

		if ((document == null) || (archiveRecord == null)) {
			final WARCReader currentReader = this.warcReader.get(usageContext);
			if (currentReader != null) {
				try {
					currentReader.close();
				} catch (final IOException ioe) {
					this.log.debug("An error happend while closing WARCReader.", ioe);
				}
			}
			throw this.noMoreResources(usageContext);
		}

		final WProcessingAnnotator wProcessingAnnotator = new WProcessingAnnotator(document);
		final DublinCoreAnnotator dublinCoreAnnotator = new DublinCoreAnnotator(document);
		try {
			wProcessingAnnotator.writeGatheringDate(this.simpleDateFormat.parse(archiveRecord.getHeader().getDate()));
		} catch (final ParseException pe) {
			this.log.warn("An error occurred when parsing the Date of the Record.", pe);
		}
		dublinCoreAnnotator.writeSource(archiveRecord.getHeader().getUrl());
		if (!this.exposedAsPrefix.isEmpty()) {
			wProcessingAnnotator.writeExposedAs(this.exposedAsPrefix + archiveRecord.getHeader().getUrl());
		}

		dublinCoreAnnotator.writeFormat(contentType);

		return document;


	}


	protected Document generateDocumentWithNative(final InputStream inputStream) {
		final Document document = ResourceFactory.createResource("warcCreator", System.nanoTime() + "", Document.class);
		try {
			this.contentManager.writeNativeContent(inputStream, document);
		} catch (final WebLabCheckedException e) {
			this.log.error(e, e);
		}
		return document;
	}


	protected String skipHttpHeader(final InputStream inputStream) throws IOException {
		final StringBuilder stringBuilder = new StringBuilder();
		int c = -1;

		while ((c = inputStream.read()) != -1) {
			stringBuilder.append((char) c);
			if (c == '\r') {
				c = inputStream.read();
				stringBuilder.append((char) c);
				if (c == '\n') {
					c = inputStream.read();
					stringBuilder.append((char) c);
					if (c == '\r') {
						c = inputStream.read();
						stringBuilder.append((char) c);
						if (c == '\n') {
							return stringBuilder.toString();
						}
					}
				}
			}
		}
		throw new IOException("Unable to detect HTTP header.");
	}


	@Override
	protected boolean validateBean(final FolderListenerConfigBean bean) {
		if (!super.validateBean(bean)) {
			return false;
		}
		if (!(bean instanceof WarcConfigBean)) {
			this.log.error("Configuration bean is not a WarcConfigBean.");
			return false;
		}
		final WarcConfigBean b = (WarcConfigBean) bean;
		if (b.getAcceptedContentTypePattern() == null) {
			this.log.error("AcceptedContentTypePattern was null in Configuration bean.");
			return false;
		}
		if (b.getAcceptedMessagePattern() == null) {
			this.log.error("MessagePattern was null in Configuration bean.");
			return false;
		}
		if (b.getAcceptedHeaders() == null) {
			this.log.error("AcceptedHeaders was null in Configuration bean.");
			return false;
		}

		try {
			Pattern.compile(b.getAcceptedContentTypePattern());
		} catch (final PatternSyntaxException pse) {
			this.log.error("AcceptedContentTypePattern [" + b.getAcceptedContentTypePattern() + "] in Configuration bean was not valid.");
			return false;
		}
		try {
			Pattern.compile(b.getAcceptedMessagePattern());
		} catch (final PatternSyntaxException pse) {
			this.log.error("AcceptedMessagePattern [" + b.getAcceptedMessagePattern() + "] in Configuration bean was not valid.");
			return false;
		}
		try {
			Pattern.compile(b.getAcceptedHeaders(), Pattern.DOTALL);
		} catch (final PatternSyntaxException pse) {
			this.log.error("AcceptedHeaders [" + b.getAcceptedHeaders() + "] in Configuration bean was not valid.");
			return false;
		}

		return true;
	}

}
