/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 * 
 * Copyright (C) 2004 - 2012 CASSIDIAN
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */

package org.ow2.weblab.webservices.listener.tool;

import java.io.File;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.regex.Pattern;

import org.apache.commons.io.monitor.FileAlterationMonitor;
import org.apache.commons.io.monitor.FileAlterationObserver;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.ow2.weblab.webservices.listener.bean.FolderListenerConfigBean;

/**
 * 
 * Encapsulate FileAlterationMonitor (from commons IO) management.
 * 
 * @author Jérémie Doucy, CASSIDIAN, an EADS company
 */
public class FolderListenerManager {


	/**
	 * monitor already started ?
	 */
	protected boolean started;


	/**
	 * FIFO to manage URL to return (First in First out).
	 */
	protected final Queue<File> fifo;


	/**
	 * Pattern used to match new files, to choose if we add it to the FIFO.
	 */
	protected final Pattern fileUrlPattern;


	/**
	 * File monitor to listen to folders.
	 */
	protected final FileAlterationMonitor monitor;


	protected final Log log;


	/**
	 * Initialises the singleton using the configuration bean.
	 * 
	 * @see FolderListenerManager#FolderListenerManager(Pattern, int, List, boolean)
	 * @param config
	 *            the bean which contains the configuration
	 */
	public FolderListenerManager(final FolderListenerConfigBean config) {
		this(Pattern.compile(config.getFileUrlPattern()), config.getUpdateInterval(), config.getFolders(), config.isRecursive(), config.isOneShotCrawler());
	}


	public FolderListenerManager(final Pattern fileUrlPattern, final int updateInterval, final List<File> folders, final boolean recursive) {
		this(fileUrlPattern, updateInterval, folders, recursive, false);
	}


	public FolderListenerManager(final Pattern fileUrlPattern, final int updateInterval, final List<File> folders, final boolean recursive,
			final boolean oneShotCrawler) {
		this.fileUrlPattern = fileUrlPattern;
		this.monitor = new FileAlterationMonitor(updateInterval);
		this.fifo = new LinkedList<File>();
		this.started = false;
		this.log = LogFactory.getLog(this.getClass());

		/*
		 * create one observer per folder
		 */
		for (final File folder : folders) {

			/*
			 * add the FIFO listener to the observer if not oneShot
			 */
			if (!oneShotCrawler) {
				final FileAlterationObserver observer = new FileAlterationObserver(folder);
				observer.addListener(new FifoFileListener(this.fifo, fileUrlPattern));
				this.monitor.addObserver(observer);
			}
			this.addFilesToFifo(folder, recursive, true);
		}
		this.log.debug("Observer initialised on folders: " + folders.toString() + ".");
	}


	protected void addFilesToFifo(final File file, final boolean recursive, final boolean topLevelFolder) {
		/*
		 * it's a file, adding it to the FIFO
		 */
		if (file.isFile()) {
			if (this.fileUrlPattern.matcher(file.getAbsolutePath()).matches()) {
				this.log.debug("Adding " + file.getAbsolutePath() + " to the FIFO.");
				this.fifo.add(file);
			} else {
				this.log.debug("File " + file.getAbsolutePath() + " not added to the FIFO due to none pattern match: " + this.fileUrlPattern);
			}
		} else if (file.isDirectory() && (recursive || topLevelFolder)) {
			/*
			 * it's a folder, recursive call
			 */
			for (final File curFile : file.listFiles()) {
				this.addFilesToFifo(curFile, recursive, false);
			}
		}
	}


	/**
	 * Recursive method which adds files to the FIFO and recall if its a folder.
	 * 
	 * @param file
	 *            the folder or file
	 * @param recursive
	 *            if true, add sub folders
	 */
	@Deprecated
	protected void addFilesToFifo(final File file, final boolean recursive) {
		this.addFilesToFifo(file, recursive, false);
	}


	/**
	 * Start monitoring on defined folders.
	 * 
	 * @return true if monitor has been started
	 */
	public synchronized boolean start() {
		if (!this.started) {
			try {
				this.monitor.start();
				this.started = true;
			} catch (final Exception e) {
				this.log.error("Error during monitor start.", e);
			}
		}
		return this.started;
	}


	/**
	 * Stop monitoring on defined folders.
	 */
	public void stop() {
		if (this.started) {
			try {
				this.monitor.stop();
				this.started = false;
			} catch (final Exception e) {
				this.log.warn("Error during monitor stop.", e);
			}
		}
	}


	/**
	 * Destroy singleton to enable new initialization.
	 */
	public synchronized void destroy() {
		this.stop();
	}


	/**
	 * FIFO which is dynamically updated.
	 * 
	 * @return the FIFO
	 */
	public Queue<File> getFifo() {
		return this.fifo;
	}

}
